/*
* tclioc.cc - a Tcl shell for iocCore
*
* Author: Dirk Zimoch
*
* DISCLAIMER: Use at your own risc and so on. No warranty, no refund.
*
* Part of this file is a modified copy of $EPICS_BASE/src/iocsh/iocsh.cpp
* For those parts, note the copyright below.
*
* Other parts have been rewritten for use with TCL
*/
#ifndef _GNU_SOURCE
// for open_memstream() function
#define _GNU_SOURCE
#endif

/*************************************************************************/
/* Copy from $EPICS_BASE/src/iocsh/iocsh.cpp starts here */

/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
/* iocsh.cpp */
/* Author:  Marty Kraimer Date: 27APR2000 */
/* Heavily modified by Eric Norum   Date: 03MAY2000 */
/* Adapted to C++ by Eric Norum   Date: 18DEC2000 */

#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <tcl.h>
#include <unistd.h>

#include "errlog.h"
#include "dbAccess.h"
#include "macLib.h"
#include "epicsStdio.h"
#include "epicsString.h"
#include "epicsStdlib.h"
#include "epicsThread.h"
#include "epicsMutex.h"
#include "envDefs.h"
#include "registry.h"
#include "epicsReadline.h"
#include "cantProceed.h"
#define epicsExportSharedSymbols
#include "iocsh.h"

#if (TCL_MAJOR_VERSION>8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION>=4))
#define TCL_STRING CONST char *
#else
#define TCL_STRING char *
#endif

extern "C" {

/*
 * Global link to pdbbase
 */
epicsShareDef struct dbBase **iocshPpdbbase;

/*
 * File-local information
 */
struct iocshCommand {
    iocshFuncDef const     *pFuncDef;
    iocshCallFunc          func;
    struct iocshCommand    *next;
};
static struct iocshCommand *iocshCommandHead;
static char iocshCmdID[] = "iocshCmd";
struct iocshVariable {
    iocshVarDef const     *pVarDef;
    struct iocshVariable  *next;
};
static struct iocshVariable *iocshVariableHead;
static char iocshVarID[] = "iocshVar";
static epicsMutexId iocshTableMutex;
static epicsThreadOnceId iocshTableOnceId = EPICS_THREAD_ONCE_INIT;

struct interpList {
    Tcl_Interp        *interp;
    struct interpList *next;
};

struct interpList* interpreters = NULL;
char* Tclioc_VariableTracer (
    ClientData clientData,
    Tcl_Interp *interp,
    TCL_STRING name1, TCL_STRING name2,
    int flags);
    
int Tclioc_CommandWrapper (
    ClientData clientData,
    Tcl_Interp *interp,
    int argc, TCL_STRING argv[]);

/*
 * I/O redirection
 */
#define NREDIRECTS   5
struct iocshRedirect {
    TCL_STRING name;
    const char *mode;
    FILE       *fp;
    FILE      *oldFp;
#ifdef __USE_GNU
    char   *ptr;
    size_t size;
#endif
};

/*
 * Set up command table mutex
 */
static void iocshTableOnce (void *)
{
    iocshTableMutex = epicsMutexMustCreate ();
}

/*
 * Lock the table mutex
 */
static void
iocshTableLock (void)
{
    epicsThreadOnce (&iocshTableOnceId, iocshTableOnce, NULL);
    epicsMutexMustLock (iocshTableMutex);
}

/*
 * Unlock the table mutex
 */
static void
iocshTableUnlock (void)
{
    epicsThreadOnce (&iocshTableOnceId, iocshTableOnce, NULL);
    epicsMutexUnlock (iocshTableMutex);
}

/*
 * Register a command
 */
void epicsShareAPI iocshRegister (const iocshFuncDef *piocshFuncDef, iocshCallFunc func)
{
    struct iocshCommand *l, *p, *n;
    int i;
    struct interpList* interplist;

    iocshTableLock ();
    for (l = NULL, p = iocshCommandHead ; p != NULL ; l = p, p = p->next) {
        i = strcmp (piocshFuncDef->name, p->pFuncDef->name);
        if (i == 0) {
            p->pFuncDef = piocshFuncDef;
            p->func = func;
            iocshTableUnlock ();
            return;
        }
        if (i < 0)
            break;
    }
    n = (struct iocshCommand *)callocMustSucceed (1, sizeof *n, "iocshRegister");
    if (!registryAdd(iocshCmdID, piocshFuncDef->name, (void *)n)) {
        free (n);
        iocshTableUnlock ();
        errlogPrintf ("iocshRegister failed to add %s\n", piocshFuncDef->name);
        return;
    }
    if (l == NULL) {
        n->next = iocshCommandHead;
        iocshCommandHead = n;
    }
    else {
        n->next = l->next;
        l->next = n;
    }
    n->pFuncDef = piocshFuncDef;
    n->func = func;
    // register command in all running TCL interpreters
    // (For those functions that register after Tclioc_Init)
    for (interplist = interpreters; interplist; interplist = interplist->next)
    {
        Tcl_CreateCommand(interplist->interp, (char*) piocshFuncDef->name,
            Tclioc_CommandWrapper, (ClientData) n, NULL);
    }
    iocshTableUnlock ();
}

/*
 * Register variable(s)
 */

void epicsShareAPI iocshRegisterVariable (const iocshVarDef *piocshVarDef)
{
    struct iocshVariable *l, *p, *n;
    int i;
    int found;
    struct interpList* interplist;

    iocshTableLock ();
    while ((piocshVarDef != NULL)
        && (piocshVarDef->name != NULL)
        && (*piocshVarDef->name != '\0')) {
        found = 0;
        for (l = NULL, p = iocshVariableHead ; p != NULL ; l = p, p = p->next) {
            i = strcmp (piocshVarDef->name, p->pVarDef->name);
            if (i == 0) {
                errlogPrintf("Warning -- iocshRegisterVariable redefining %s.\n", piocshVarDef->name);
                p->pVarDef = piocshVarDef;
                found = 1;
                break;
            }
            if (i < 0)
                break;
        }
        if (!found) {
            n = (struct iocshVariable *)callocMustSucceed(1, sizeof *n, "iocshRegisterVariable");
            if (!registryAdd(iocshVarID, piocshVarDef->name, (void *)n)) {
                free(n);
                iocshTableUnlock();
                errlogPrintf("iocshRegisterVariable failed to add %s.\n", piocshVarDef->name);
                return;
            }
            if (l == NULL) {
                n->next = iocshVariableHead;
                iocshVariableHead = n;
            }
            else {
                n->next = l->next;
                l->next = n;
            }
            n->pVarDef = piocshVarDef;
            
            // register variable in all running TCL interpreters
            // (This is most probably only the main interpreter)
            for (interplist = interpreters; interplist; interplist = interplist->next)
            {
                /* use tracer to create TCL variable */
                Tclioc_VariableTracer ((ClientData) piocshVarDef, interplist->interp,
                    (char*) piocshVarDef->name, NULL,
                    TCL_GLOBAL_ONLY | TCL_TRACE_READS);
                /* install tracer to redirect read and write to EPICS variable */
                Tcl_TraceVar(interplist->interp, (char*) piocshVarDef->name,
                    TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS | TCL_GLOBAL_ONLY,
                    Tclioc_VariableTracer, (ClientData) piocshVarDef);
            }
        }
        piocshVarDef++;
    }
    iocshTableUnlock ();
}

/*
 * Free storage created by iocshRegister/iocshRegisterVariable
 */
void epicsShareAPI iocshFree(void) 
{
    struct iocshCommand *pc, *nc;
    struct iocshVariable *pv, *nv;

    iocshTableLock ();
    for (pc = iocshCommandHead ; pc != NULL ; ) {
        nc = pc->next;
        free (pc);
        pc = nc;
    }
    for (pv = iocshVariableHead ; pv != NULL ; ) {
        nv = pv->next;
        free (pv);
        pv = nv;
    }
    iocshTableUnlock ();
}

/* Redirection functions have been modified for use with TCL */

/*
 * Open redirected I/O
 */

static int
openRedirect(Tcl_Interp *interp, struct iocshRedirect *redirect)
{
    int i;

    for (i = 0 ; i < NREDIRECTS ; i++, redirect++) {
        if (redirect->name != NULL) {
            if (redirect->mode[0] == '@') {
                // redirect to an already open file:
                // create a copy of an existing file pointer
                // that can be closed independently of the
                // existing open file pointer
                FILE* fp;
                if (Tcl_GetOpenFile(interp, redirect->name, 1, 1,
                    (ClientData*) &fp) != TCL_OK) {
                    redirect->fp = NULL;
                } else {
                    Tcl_Flush(Tcl_GetChannel(interp, redirect->name, NULL));
                    redirect->fp = fdopen(dup(fileno(fp)), redirect->mode+1);
                }
            }
#ifdef __USE_GNU
            else if (redirect->mode[0] == '$') {
                // write to an variable:
                // create a memstream
                redirect->fp = open_memstream(&redirect->ptr, &redirect->size);
                if (redirect->fp == NULL) {
                    Tcl_ResetResult(interp);
                    Tcl_AppendResult(interp,
                        "Can't open memstream: ", strerror(errno), NULL);
                }
            }
#endif
            else {
                redirect->fp = fopen(redirect->name, redirect->mode);
                if (redirect->fp == NULL) {
                    Tcl_ResetResult(interp);
                    Tcl_AppendResult(interp,
                        "Can't open \"", redirect->name, "\": ",
                        strerror(errno), NULL);
                }
            }
            if (redirect->fp == NULL) {
                while (i--) {
                    redirect--;
                    if (redirect->fp) {
                        fclose(redirect->fp);
                        redirect->fp = NULL;
#ifdef __USE_GNU
                        if (redirect->ptr)
                        {
                            free(redirect->ptr);
                            redirect->ptr = NULL;
                        }
#endif
                    }
                }
                return -1;
            }
        }
    }
    return 0;
}

/*
 * Start I/O redirection
 */
static void
startRedirect(Tcl_Interp *interp, struct iocshRedirect *redirect)
{
    int i;

    for (i = 0 ; i < NREDIRECTS ; i++, redirect++) {
        if (redirect->fp != NULL) {
            switch(i) {
            case 0:
                redirect->oldFp = epicsGetThreadStdin();
                epicsSetThreadStdin(redirect->fp);
                break;
            case 1:
                redirect->oldFp = epicsGetThreadStdout();
                epicsSetThreadStdout(redirect->fp);
                break;
            case 2:
                redirect->oldFp = epicsGetThreadStderr();
                epicsSetThreadStderr(redirect->fp);
                break;
            }
        }
    }
}

/*
 * Finish up I/O redirection
 */
static void
stopRedirect(Tcl_Interp *interp, struct iocshRedirect *redirect)
{
    int i;

    for (i = 0 ; i < NREDIRECTS ; i++, redirect++) {
        if (redirect->fp != NULL) {
            if (fclose(redirect->fp) != 0)
            {
                Tcl_ResetResult(interp);
                Tcl_AppendResult(interp,
                    "Error closing \"", redirect->name, "\": ",
                    strerror(errno));
            }
            redirect->fp = NULL;
            switch(i) {
            case 0: epicsSetThreadStdin(redirect->oldFp);  break;
            case 1: epicsSetThreadStdout(redirect->oldFp); break;
            case 2: epicsSetThreadStderr(redirect->oldFp); break;
            }
#ifdef __USE_GNU
            if (redirect->mode[0] == '$') {
                if (redirect->size && redirect->ptr[redirect->size-1] == '\n') {
                    redirect->ptr[redirect->size-1] = '\0';
                }
                if (redirect->mode[1] == '$') {
                    // redirect to result
                    Tcl_SetResult(interp, redirect->ptr, (Tcl_FreeProc *) free);
                }
                else
                {
                    // redirect to variable
                    Tcl_SetVar(interp, redirect->name, redirect->ptr,
                        (redirect->mode[1] == '+' ? TCL_APPEND_VALUE : 0) |
                        TCL_LEAVE_ERR_MSG);
                    free(redirect->ptr);
                }
            }
#endif
        }
        redirect->name = NULL;
    }
}

/*
 * "help" command
 */
static const iocshArg helpArg0 = { "[command ...]",iocshArgArgv};
static const iocshArg *helpArgs[1] = {&helpArg0};
static const iocshFuncDef helpFuncDef =
    {"help",1,helpArgs};
static void helpCallFunc(const iocshArgBuf *args)
{
    int argc = args[0].aval.ac;
    const char * const * argv = args[0].aval.av;
    struct iocshFuncDef const *piocshFuncDef;
    struct iocshCommand *pcmd;

    if (argc == 1) {
        int l, col = 0;

        printf ("Type `help command_name' to get more information about a particular command.\n");
        iocshTableLock ();
        for (pcmd = iocshCommandHead ; pcmd != NULL ; pcmd = pcmd->next) {
            piocshFuncDef = pcmd->pFuncDef;
            l = strlen (piocshFuncDef->name);
            if ((l + col) >= 79) {
                fputc ('\n', stdout);
                col = 0;
            }
            fputs (piocshFuncDef->name, stdout);
            col += l;
            if (col >= 64) {
                fputc ('\n', stdout);
                col = 0;
            }
            else {
                do {
                    fputc (' ', stdout);
                    col++;
                } while ((col % 16) != 0);
            }
        }
        if (col)
            fputc ('\n', stdout);
        iocshTableUnlock ();
    }
    else {
        for (int iarg = 1 ; iarg < argc ; iarg++) {
            for (pcmd = iocshCommandHead ; pcmd != NULL ; pcmd = pcmd->next) {
                piocshFuncDef = pcmd->pFuncDef;
                if (epicsStrGlobMatch(piocshFuncDef->name, argv[iarg]) != 0) {
                    fputs (piocshFuncDef->name, stdout);
                    for (int a = 0 ; a < piocshFuncDef->nargs ; a++) {
                        const char *cp = piocshFuncDef->arg[a]->name;
                        if ((piocshFuncDef->arg[a]->type == iocshArgArgv)
                         || (strchr (cp, ' ') == NULL)) {
                            fprintf (stdout, " %s", cp);
                        }
                        else {
                            fprintf (stdout, " '%s'", cp);
                        }
                    }
                    fprintf (stdout,"\n");;
                }
            }
        }
    }
}

/* Copy from $EPICS_BASE/src/iocsh/iocsh.cpp ends here */
/*************************************************************************/

int Tclioc_CommandWrapper (
    ClientData clientData,
    Tcl_Interp *interp,
    int argc, TCL_STRING argv[])
{
    struct iocshCommand* piocshCommand = (struct iocshCommand*) clientData;
    iocshArgBuf *argBuf = NULL;
    char** EPICSargv = NULL;
    int i, n;
    int wasOkToBlock;
    struct iocshRedirect redirects [NREDIRECTS];
    int fd, ic;
    char c;
    
    // convert all arguments from TCL to EPICS
    // missing arguments are set to default values
    // surplus arguments are ignored
    // redirections <, >, n>, >>, n>>, >@, n>@ supported
    // only with GNU compiler:
    //    >$, >>$, n>$, n>>$ redirect to a variable
    //    >$$, n>$$ redirect to command result
    
    bzero(redirects, sizeof (redirects));
    
    // allocate array of pointers to all arguments
    argBuf = (iocshArgBuf*) Tcl_Alloc(
        piocshCommand->pFuncDef->nargs * sizeof(iocshArgBuf));
    bzero(argBuf,piocshCommand->pFuncDef->nargs * sizeof(iocshArgBuf));
    if (!argBuf)
    {
        Tcl_SetResult (interp, "Out of memory!", TCL_STATIC);
        return TCL_ERROR;
    }
    
    // argv[0] contains command    
    // TCL argc counts command and args, i.e. i = 1 ... argc-1
    // EPICS nargs counts only args, i.e. n = 0 ... nargs-1
    
    for (i = 1, n = 0; i < argc || n < piocshCommand->pFuncDef->nargs; i++)
    {
        if (i < argc)
        {
            // look for i/o redirection
            ic = 0;
            c = argv[i][0];
            fd = 1;
            if (c >= '1' && c < '0'+NREDIRECTS && argv[i][ic+1] == '>')
            {
                // output redirection of numbered fd
                ic++;
                fd = c - '0';
                c = '>';
            }
            if (c == '>')
            {
                // output redirection
                ic++;
                if (argv[i][ic] == '@')
                {
                    // >@ fileId   write to already open file/pipe/socket
                    redirects[fd].mode = "@w";   
                    ic++;
                }
                else
                if (argv[i][ic] == '$')
                {
#ifdef __USE_GNU
                    // >$ variable   write to variable
                    redirects[fd].mode = "$";   
                    ic++;
                    if (argv[i][ic] == '$')
                    {
                        // >$$   write to result
                        redirects[fd].mode = "$$";
                        redirects[fd].name = "-";
                        continue; // no redirection argument
                    }
#else
                    Tcl_SetResult (interp, "Redirection >$ unsupported", TCL_STATIC);
                    goto fault; 
#endif
                }
                else
                if (argv[i][ic] == '>')
                {
                    // >> filename   append output
                    redirects[fd].mode = "a";
                    ic++;
#ifdef __USE_GNU
                    if (argv[i][ic] == '$')
                    {
                        // >>$ variable  append to variable
                        redirects[fd].mode = "$+";   
                        ic++;
                    }
#else
                    Tcl_SetResult (interp, "Redirection >>$ unsupported", TCL_STATIC);
                    goto fault; 
#endif
                }
                else
                {
                    // > filename   overwrite output
                    redirects[fd].mode = "w";
                }
            }
            if (c == '<')
            {
                // input redirection
                ic++;
                fd = 0;
                if (argv[i][ic] == '@')
                {
                    // <@ fileId   read from already open file/pipe/socket
                    redirects[fd].mode = "@r";   
                    ic++;
                } else {
                    // < filename   read from file
                    redirects[fd].mode = "r";
                }   
            }
            if (ic)
            {
                // redirection found
                if (!argv[i][ic])
                {
                    // channel/variable/filehandle in next argument
                    i++;
                    ic=0;
                    if (i >= argc)
                    {
                        Tcl_SetResult (interp, "missing filename", TCL_STATIC);
                        goto fault;
                    }
                }
                redirects[fd].name = &argv[i][ic];
                continue; // don't increment n
            }
        }
        
        if (n >= piocshCommand->pFuncDef->nargs)
        {
            Tcl_SetResult (interp, "too many arguments", TCL_STATIC);
            goto fault;
        }
        
        switch (piocshCommand->pFuncDef->arg[n]->type)
        {
            case iocshArgInt:
                if (i < argc)
                {
                    // convert argument to integer
                    if (Tcl_GetInt (interp, argv[i], &argBuf[n].ival) != TCL_OK)
                    {
                        goto fault;
                    }
                }
                else
                {
                    // default integer argument is 0
                    argBuf[n].ival = 0;
                }
                break;
            case iocshArgDouble:
                if (i < argc)
                {
                    // convert argument to double
                    if (Tcl_GetDouble (interp, argv[i], &argBuf[n].dval) != TCL_OK)
                    {
                        goto fault;
                    }
                }
                else
                {
                    // default double argument is 0.0
                    argBuf[n].dval = 0.0;
                }
                break;
            case iocshArgString:
                if (i < argc)
                {
                    // no need to convert string argument
                    argBuf[n].sval = (char*) argv[i];
                }
                else
                {
                    // default string argument is NULL
                    argBuf[n].sval = NULL;
                }
                break;
            case iocshArgPersistentString:
                if (i < argc)
                {
                    // copy string argument
                    argBuf[n].sval = epicsStrDup(argv[i]);
                    if (!argBuf[n].sval)
                    {
                        Tcl_SetResult (interp, "Out of memory!", TCL_STATIC);
                        goto fault;
                    }
                }
                else
                {
                    // default string argument is NULL
                    argBuf[n].sval = NULL;
                }
                break;
            case iocshArgPdbbase:
                if (i >= argc || argv[i] == NULL || argv[i][0] == '0' ||
                        strcmp(argv[i], "pdbbase") == 0)
                    // default and only valid and pdbbase argument is pdbbase
                    argBuf[n].vval = *iocshPpdbbase;
                else
                {
                    Tcl_AppendResult (interp, "Expecting pdbbase but got ",
                        argv[i], NULL);
                    goto fault;
                }
                break;
            case iocshArgArgv:
                if (i < argc)
                {
                    // store remaining arguments in argv array
                    if (!EPICSargv)
                    {
                        // 1st argv argument
                        // create array and init argBuf[n].aval
                        EPICSargv =  (char **) Tcl_Alloc((argc-i+1) * sizeof(char*));
                        argBuf[n].aval.ac = (n?0:1);
                        argBuf[n].aval.av = EPICSargv;
                        if (!n) argBuf[0].aval.av[0] = (char*) argv[0];
                    }
                    EPICSargv[argBuf[n].aval.ac++] = (char*) argv[i];
                    continue; // don't increment n any more
                }
                else
                {
                    if (EPICSargv)
                    {
                        // done with all args
                        break;
                    }
                    if (n)
                    {
                        // default argv argument is empty
                        argBuf[n].aval.ac = 0;
                        argBuf[n].aval.av = NULL;
                    }
                    else
                    {
                        // but if argv is first arg, default is command name
                        argBuf[0].aval.ac = 1;
                        argBuf[0].aval.av = (char**) argv;
                    }
                }
                break;
            default:
                Tcl_SetResult (interp, "unsupported argument type", TCL_STATIC);
                goto fault;
        }
        n++;
    }
    
    wasOkToBlock = epicsThreadIsOkToBlock();
    epicsThreadSetOkToBlock(1);
    
    if (openRedirect(interp, redirects) < 0) goto fault;
    startRedirect(interp, redirects);
    piocshCommand->func(argBuf);
    stopRedirect(interp, redirects);
    
    epicsThreadSetOkToBlock( wasOkToBlock);
    Tcl_Free((char*)EPICSargv);
    Tcl_Free((char*)argBuf); 
    return TCL_OK;
fault:
    Tcl_Free((char*)EPICSargv);
    Tcl_Free((char*)argBuf); 
    return TCL_ERROR;
}

char* Tclioc_VariableTracer (
    ClientData clientData,
    Tcl_Interp *interp,
    TCL_STRING name1, TCL_STRING name2,
    int flags)
{
    iocshVarDef* pVarDef = (iocshVarDef*) clientData;
    char buffer[TCL_DOUBLE_SPACE];
    TCL_STRING sval;
    
    if (flags & (TCL_TRACE_READS|TCL_TRACE_UNSETS))
    {
        // unset is not allowed for EPICS variables -> recreate TCL variable
        switch (pVarDef->type)
        {
            case iocshArgInt:
                sprintf (buffer, "%i", *(int*)pVarDef->pval);
                sval = buffer;
                break;
            case iocshArgDouble:
                Tcl_PrintDouble (interp, *(double*)pVarDef->pval, buffer);
                sval = buffer;
                break;
            case iocshArgString:
                sval = (char*) pVarDef->pval;
                break;
            case iocshArgPdbbase:
                return "unsupported iocsh variable type iocshArgPdbbase";
            case iocshArgArgv:
                return "unsupported iocsh variable type iocshArgArgv";
            case iocshArgPersistentString:
                return "unsupported iocsh variable type iocshArgPersistentString";
            default:
                return "unsupported iocsh variable type (unknown)";
        }
        Tcl_SetVar2(interp, name1, name2, sval,
            (flags & TCL_GLOBAL_ONLY)|TCL_LEAVE_ERR_MSG);
        if (flags & TCL_TRACE_UNSETS)
        {
            /* reinstall tracer*/
            Tcl_TraceVar(interp, name1,
                TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS | TCL_GLOBAL_ONLY,
                Tclioc_VariableTracer, clientData);
        }
        return NULL;
    }
    if (flags & TCL_TRACE_WRITES)
    {
        sval = Tcl_GetVar2(interp, name1, name2, flags & TCL_GLOBAL_ONLY);
        switch (pVarDef->type)
        {
            case iocshArgInt:
                if (Tcl_GetInt (interp, sval, (int*)pVarDef->pval) == TCL_ERROR)
                {
                    return interp->result;
                }
                break;
            case iocshArgDouble:
                if (Tcl_GetDouble (interp, sval, (double*)pVarDef->pval) == TCL_ERROR)
                {
                    return interp->result;
                }
                break;
            case iocshArgString:
                pVarDef->pval = (void*) sval;
                break;
            case iocshArgPdbbase:
                return "unsupported iocsh variable type iocshArgPdbbase";
            case iocshArgArgv:
                return "unsupported iocsh variable type iocshArgArgv";
            case iocshArgPersistentString:
                return "unsupported iocsh variable type iocshArgPersistentString";
            default:
                return "unsupported iocsh variable type (unknown)";
        }
        return NULL;
    }
    return NULL;
}

int Tclioc_Init(Tcl_Interp* interp)
{
    struct iocshCommand* cmd;
    struct iocshVariable* var;
    struct interpList** prev;
    
    iocshRegister(&helpFuncDef,helpCallFunc);
    iocshTableLock ();
    // create a tcl command for each EPICS command
    for (cmd = iocshCommandHead; cmd != NULL; cmd = cmd->next)
    {
        Tcl_CreateCommand(interp, (char*) cmd->pFuncDef->name,
            Tclioc_CommandWrapper, (ClientData) cmd, NULL);
    }
    iocshTableUnlock ();
    
    // trace a global TCL variable for each already registered EPICS variable
    // (This is none for the main interpreter, see iocshRegisterVariable)        
    for (var = iocshVariableHead; var != NULL; var = var->next)
    {
        // use tracer to create TCL variable
        Tclioc_VariableTracer ((ClientData) var->pVarDef, interp,
            (char*) var->pVarDef->name, NULL,
            TCL_GLOBAL_ONLY | TCL_TRACE_READS);
        // install tracer to redirect read and write to EPICS variable
        Tcl_TraceVar(interp, (char*) var->pVarDef->name,
            TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS | TCL_GLOBAL_ONLY,
            Tclioc_VariableTracer, (ClientData) var->pVarDef);
    }
    
    // hook this interpreter in list for iocshRegisterVariable
    for (prev = &interpreters; *prev; prev=&(*prev)->next);
    *prev = new interpList;
    (*prev)->interp = interp;
    (*prev)->next = NULL;

    return TCL_OK;
}

} /* extern "C" */
